/******************************************************************************
  REPSYS replace_syscall.c
  ========================
  versione    01
  autore      Calzoni Pietro aka Calzo <calzog@gmail.com>
                LUGMan - Linux Users Group Mantova
                www.lugman.org
  licenza     GPL
  Kernel      2.6
  Kernel test 2.6.32.7
  System      SalckWare 13
  compilers   gcc-4.3.3
 ------------------------------------------------------------------------------
Obiettivo  realizzare la sostituzione di una chiamata di sistema. Per farlo
occorre:
- individuare la sys_call_table
- rendere quest'area di memoria scrivibile
- cambiare la chiamata che si desidera sostituire con quella nuova
- ripristinare la syscall in fase di unload del modulo
Il passo principale  individuare dove si trova la sys_call_table. Dal kernel 2.6
questo simbolo non  pi esportato e quindi non  visibile in /proc/kallsyms.
L'individuazione di questo puntatore pu essere fatto in diversi modi pi o meno
complessi. In questo esempio anlizzeremo il metodo pi semplice, ossia andare a
leggere le informazioni sui puntatori che interessano direttaemnte dal file
System.map. E' chiaro quindi che questo metodo necessita di avere di sorgenti
del kernel sul quale si vuole operare.
 ------------------------------------------------------------------------------
L'esempio qui proposto prevede la sostituzione della chiamata open e cercheremo
di rendere incancellabile e immodificabile un file specifico dal filesystem.
La sys_call_table verr passata al modulo come parametro.
 ------------------------------------------------------------------------------
NOTA: Use the virt_to_page() macro to convert an address to a page  and offset.
NOTA: asm/page.h asm-generic/page.h
 ******************************************************************************/
#include <linux/init.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/version.h>

#include <asm/unistd.h>	// contiene i numeri delle chiamate __NR_xxx
#include <asm/page.h>	// per avere le funzioni per la memoria come virt_to_page (riferito ad asm-generic/page.h)
#include <asm/linkage.h>
//#include <asm/memory_model.h>	// Get pfn_to_page e page_to_pfn
#include <linux/mm.h>			// Get struct page
#include <linux/syscalls.h>		// Get syscall sys_xxx, ma non serve a molto

MODULE_LICENSE("GPL");
MODULE_VERSION("0.1");
MODULE_AUTHOR("Calzo");
MODULE_DESCRIPTION("Syscall Hijaking: sostituzione della chiamata 'getpid'");

// PARAMETERS
static unsigned long sct_addr = 0xc1319110;	// Syscall Table address - passato come paraemtro
module_param(sct_addr, ulong, S_IRUGO);    // see LDD3 cap 2, pag 36 e il file linux/stat.h 
MODULE_PARM_DESC(sct_addr, "Syscall Table address");

static unsigned long pages_rw_addr = 0xc10133da;	// Indirizzo della funzione set_pages_rw
module_param(pages_rw_addr, ulong, S_IRUGO);    // see LDD3 cap 2, pag 36 e il file linux/stat.h 
MODULE_PARM_DESC(pages_rw_addr, "Address of set_pages_rw()");

static unsigned long pages_ro_addr = 0xc10132f7;	// Indirizzo della funzione set_pages_ro
module_param(pages_ro_addr, ulong, S_IRUGO);    // see LDD3 cap 2, pag 36 e il file linux/stat.h 
MODULE_PARM_DESC(pages_rw_addr, "Address of set_pages_ro()");

static void *original_syscall;							// puntatore alla syscall originale
static unsigned long *sys_call_table;
static void (*pages_rw)(struct page *page, int numpages);
static void (*pages_ro)(struct page *page, int numpages);
struct page *sys_call_table_page;	// puntatore alla pagina

int __init repsys_init_module(void);
void repsys_clenup_module(void);
static int counter = 0;

pid_t (*old_getpid)(void);

/* sys_new_getpid  la nuova funzione per ottenere il numero del processo.
Si noti che asmlinkage serve per fare in modo che i parametri della funzione
vengano passati via stack e non via registro
*/
asmlinkage pid_t sys_new_getpid(void)
{
	printk(KERN_ALERT "getpid chiamata %d volte\n", ++counter);
	return old_getpid();
}


/* _init_module inizializza il modulo.
Subito viene valutato se l'indirizzo passato  corretto. Come noto, il kernel
alloca di default tutto ci che lo riguarda nell'ultimo giga di RAM virtuale
(almeno per sistemi x86) e quindi l'indirizzo deve essere >0xC0000000. Se 
l'indirizzo  valido non vengono fatti altri check! quindi occhio!!
Gli stessi controlli vengono fatti per gli altri indirizzi.
*/
int __init repsys_init_module(void)
{
	if(sct_addr<=0xC000000 ) {
		printk(KERN_ALERT "Invalid SysCall Table address\n");
		return -1;
	}

	if(pages_rw_addr<=0xC000000 ) {
		printk(KERN_ALERT "Invalid set_page_rw() address\n");
		return -2;
	}

	if(pages_ro_addr<=0xC000000 ) {
		printk(KERN_ALERT "Invalid set_page_ro() address\n");
		return -3;
	}
	// --------- //
	printk(KERN_ALERT "Setting sys_call_table RW...\n");
	sys_call_table = (unsigned long*)sct_addr;
	pages_rw = (void*)pages_rw_addr;
	pages_ro = (void*)pages_ro_addr;

	// Si noti che page_rw in slackware non serve, mentre dovrebbe servire in RedHat o simili.
	// Questo pech RH pone read-only la system-table, ma anche Slackware! eppure io sovrascrivo
	// tutti i puntatori che mi pare senza trasformare le pagine da ro in rw. Non capisco perch
	printk(KERN_ALERT "Ricalcolo indirizzi:\n");
	sys_call_table_page = virt_to_page(*sys_call_table);	// o anche sct_addr
	pages_ro(sys_call_table_page, 1);
	original_syscall = (void*)sys_call_table[__NR_getpid];	// salvo la vecchia chiamata
	old_getpid = original_syscall;
	printk(KERN_ALERT "   sys_call_table (virtual): 0x%X\n   sys_call_table (page): 0x%X\n   old sys_call (virtual): %X\n",
			(int)sys_call_table, (int)sys_call_table_page, (int)original_syscall);

	sys_call_table[__NR_getpid] = (unsigned long)sys_new_getpid;
	
//	printk(KERN_INFO "__pa(x)=%x\n__va(x)=%x\n", (int)__pa(sct_addr),(int)__va(__pa(sct_addr)));
	
	return 0;
}

void repsys_clenup_module()
{
	printk(KERN_INFO "Ripristino chiamata di sistema originale.\n");
	pages_ro(sys_call_table_page, 1);						// se era tornata ro la rimetto rw
	sys_call_table[__NR_getpid] = (unsigned long)original_syscall;
}

module_init(repsys_init_module);
module_exit(repsys_clenup_module);
//EXPORT_SYMBOL(sys_new_getpid);