/******************************************************************************
  REPSYS replace_syscall_idt_getpid.c
  ===================================
  versione    01
  autore      Calzoni Pietro aka Calzo <calzog@gmail.com>
                LUGMan - Linux Users Group Mantova
                www.lugman.org
  licenza     GPL
  Kernel      2.6
  Kernel test 2.6.35.7
  System      SalckWare 13
  compilers   gcc-4.3.3
 ------------------------------------------------------------------------------
Obiettivo  realizzare la sostituzione di una chiamata di sistema. Per farlo
occorre:
- individuare la sys_call_table
- rendere quest'area di memoria scrivibile
- cambiare la chiamata che si desidera sostituire con quella nuova
- ripristinare la syscall in fase di unload del modulo
Il passo principale  individuare dove si trova la sys_call_table. Dal kernel 2.6
questo simbolo non  pi esportato e quindi non  visibile in /proc/kallsyms.
L'individuazione di questo puntatore pu essere fatto in diversi modi pi o meno
complessi.
In questo modulo verr mostrato come scoprire l'indirizzo della sys_call_table
sfruttando l'istruzione SIDT (Store Interrupt Descriptor Table) che permette di
ottene due valori "limit" e "base" che permettono di ottenere il vettore degli
interrupt software (da 0 a 255 massmo). "base"  il primo valore e "limit" 
l'ultimo. Il vattore  rappresentabile in questo modo:
                      |--------------------------------|
     base ----------> |  Address int 0x0  - 8byte      |
                      |--------------------------------|
                      | [...]                          |
                      |--------------------------------|
     base+8*0x80 ---> |  Address int 0x80 - 8byte      | --> system_call() addr
                      |--------------------------------|
                      | [...]                          |
                      |--------------------------------|
     base+limit ----> |  end                           |
                      |--------------------------------|
La system_call() accede direttamente alla sys_call_table e quindi ne conosce
l'indirizo. Tale indirizzo  contenuto nella variabile puntata dai 4 byte
successivi all'opcode dell'istruzione call. Quindi i numeri esadecimali da
cercare sono:
		FF 14 85 XX XX XX XX	dove:
	FF 14 85  l'opcode della call
	XX XX XX XX punta al valore dell'array sys_call_table
 ------------------------------------------------------------------------------
In questo esempio mostreremo come intercettare e sostituire la chiamata getpid()
aggiungendo una segnalazione ogni volta che wssa viene chiamata.
 ******************************************************************************/
#include <linux/init.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/version.h>

#include <asm/unistd.h>	// contiene i numeri delle chiamate __NR_xxx
#include <asm/page.h>	// per avere le funzioni per la memoria come virt_to_page (riferito ad asm-generic/page.h)
#include <asm/linkage.h>
//#include <asm/memory_model.h>	// Get pfn_to_page e page_to_pfn
#include <linux/mm.h>			// Get struct page
#include <linux/syscalls.h>		// Get syscall sys_xxx
#include <linux/mm_types.h>

//#include <linux/kernel.h>
//#include <linux/file.h>
//#include <linux/fs.h>

#define MAX_SEEK	500

MODULE_LICENSE("GPL");
MODULE_VERSION("0.1");
MODULE_AUTHOR("Calzo");
MODULE_DESCRIPTION("Detect sys_call_table uging IDT");

// Struttura ottenuta tramite l'istruzione "sidt" per sistemi a 32bit
struct {
        unsigned short limit;
        unsigned int base;
} __attribute__ ((packed)) idt_struct;

//struttura che permette l'individuazione delle funzioni assegnate agli
// interrupt. Nel nostro caso estrarremo la funzione legata all'interrupt 0x80
struct {
        unsigned short off1;
        unsigned short sel;
        unsigned char none,flags;
        unsigned short off2;
} __attribute__ ((packed)) idt;

static unsigned long *sys_call_table;
static unsigned long system_call_ptr;
static int counter=0;

int __init repsysidt_init_module(void);
void repsysidt_cleanup_module(void);

// funzioni di (s)protezione della memoria
//static void (*pages_rw)(struct page *page, int numpages);
//static void (*pages_ro)(struct page *page, int numpages);

// prototipo per le funzioni di sistema da sostituire
asmlinkage pid_t	(*old_getpid)(void);

/* sys_new_getpid  la nuova funzione per ottenere il numero del processo.
Si noti che asmlinkage serve per fare in modo che i parametri della funzione
vengano passati via stack e non via registro
*/
asmlinkage pid_t sys_new_getpid(void)
{
	printk(KERN_ALERT "getpid chiamata %d volte\n", ++counter);
	return old_getpid();
}


/* _findout_sct_sddr cerca la sys_call_table
Il trucco consiste nell'estrapolare il puntatore che contiene l'indirizzo alla sys_call_table.
Passando l'indirizzo del codice della system_call() possiamo andare ad elaborarlo estraendo
il puntatore usato dall'istruzione call *sys_call_table(,%eax,4) che si trova in entry_32.S.
L'opcode di questa funzione sar FF 14 85 XX XX XX XX e ci dovremo spostare subito dopo il
valore 0x85 (ossia dovremo puntare al primo XX).
	system_call_addr : indirizzo del codice della funzione system_call()
*/
void *repsysidt_findout_sct_addr(void *system_call_addr)	// scopre l'indirizzo della system_call_table
{
	unsigned char *p;	// puntatore generico accessibile byte per byte
	int i;
	
	p = (unsigned char*)system_call_addr;	// punto all'indirizzo del codice della funzione
	for(i=0; (!(p[0]==0xFF && p[1]==0x14 && p[2]==0x85)) && i<MAX_SEEK; p++, i++); //sposto il puntatore alla chiamata syscall_call

	if(i<MAX_SEEK)	// se non abbiamo sforato adesso leggiamo dove punta il puntatore, ossia scoviamo la sys_call_table
	{
		p += 3;		// avanzo nel codice per selezionare il puntatore
		return (void *)*(unsigned long*)p;	// questo  l'indirizzo della sys_call_table
	}

	return NULL;	// qui non ci devo arrivare
}


/* _init_module inizializza il modulo.
La funzione  strutturata nelle seguenti parti:
1- sidt: otteniamo il vettore degli interrupt
2- calcoliamo l'indirizzo della funzione system_call()
3- otteniamo l'indirizzo dell'array sys_call_table
4- memoriziamo le vecchie chiamate di
5- sostituiamo le chiamate di sistema con le nostre
*/
int __init repsysidt_init_module(void)
{
	asm("sidt %0" : "=m" (idt_struct));
	memcpy(&idt, (void*)(idt_struct.base+8*0x80), sizeof(idt));
	system_call_ptr = (idt.off2 << 16) + idt.off1;

	printk("system_call() address: 0x%X\n", (int)system_call_ptr);
	
	// a questo punto si cerca la sys_call_table
	sys_call_table = (unsigned long*)repsysidt_findout_sct_addr((void*)system_call_ptr);

	if(!sys_call_table)
	{
		printk(KERN_WARNING "sys_call_table address not found!\n");
		return -1;
	}

	printk("system_call() address: 0x%X\n", (int)system_call_ptr);
	
	// a questo punto si cerca la sys_call_table
	printk("sys_call_table address: 0x%X\n", (int)sys_call_table);
	printk("sys_getpid     address: 0x%X\n", (int)(sys_call_table[__NR_getpid]));

	old_getpid	= (void*)sys_call_table[__NR_getpid];
	sys_call_table[__NR_getpid]	= (unsigned long)sys_new_getpid;

	printk(KERN_ALERT "Sostituzione sys_getpid - address: 0x%X\n", (int)(sys_call_table[__NR_getpid]));

	return 0;
}

/*  _cleanup_module ripristina le vecchie chiamate di sistema
*/
void repsysidt_cleanup_module()
{
	sys_call_table[__NR_getpid]=(unsigned long)old_getpid;
	printk(KERN_INFO "Ripristino syscall originale\n");
}

module_init(repsysidt_init_module);
module_exit(repsysidt_cleanup_module);
